import $$observable from 'symbol-observable';

var Core, cancelTask, comparer, processUpdates;

comparer = function(v, k, b, is_array, path, r) {
  var new_path, ref, ref1;
  new_path = path.concat([k]);
  if (is_array && !(((v != null ? v.id : void 0) && (v != null ? v.id : void 0) === ((ref = b[k]) != null ? ref.id : void 0)) || ((v != null ? v._id : void 0) && (v != null ? v._id : void 0) === ((ref1 = b[k]) != null ? ref1._id : void 0))) || !((v != null) && ((b != null ? b[k] : void 0) != null) && (v instanceof Object))) {
    return r.push(new_path.concat([v]));
  }
  return r.push.apply(r, diff(v, b[k], new_path));
};

cancelTask = function(handle, defer) {
  var index, q;
  q = defer ? Core.queues.deferred : Core.queues.current;
  index = handle - (q.nextHandle - q.tasks.length);
  if (q.nextIndex <= index) {
    return q.tasks[index] = null;
  }
};

processUpdates = function(q, current) {
  var count, err, mark, task;
  count = 0;
  mark = 0;
  while (q.nextIndex < q.tasks.length) {
    if (!(task = q.tasks[q.nextIndex])) {
      q.nextIndex++;
      continue;
    }
    if (current && task.defer) {
      task.handle = void 0;
      queueTask(task, task.value);
      q.nextIndex++;
      continue;
    }
    if (q.nextIndex > mark) {
      if (count++ > 5000) {
        console.error('Exceeded max task recursion');
        q.nextIndex = 0;
        return q.tasks = [];
      }
      mark = q.tasks.length;
    }
    try {
      task(task.value);
      task.handle = task.dhandle = task.value = void 0;
    } catch (error) {
      err = error;
      console.error(err);
    }
    q.nextIndex++;
  }
  q.nextIndex = 0;
  q.tasks = [];
};

var Core$1 = Core = {
  context: null,
  frozen: false,
  queues: {
    deferred: {
      tasks: [],
      nextIndex: 0,
      nextHandle: 1
    }
  },
  clock: 0
};

var queueTask = function(task, value) {
  var q;
  if (Core.frozen) {
    if (task.handle != null) {
      cancelTask(task.handle);
    }
    Core.queues.current.tasks.push(task);
    task.value = value;
    task.handle = Core.queues.current.nextHandle++;
    return;
  }
  if (task.defer) {
    if (task.dhandle != null) {
      cancelTask(task.dhandle, task.defer);
    }
    q = Core.queues.deferred;
    if (!q.tasks.length) {
      Promise.resolve().then(function() {
        processUpdates(q);
        return Core.clock++;
      });
    }
    q.tasks.push(task);
    task.value = value;
    task.dhandle = q.nextHandle++;
    return;
  }
  task(value);
};

var run = function(fn) {
  var execute, prevQueue;
  if (!Core.frozen) {
    execute = true;
    prevQueue = Core.queues.current;
    Core.queues.current = {
      tasks: [],
      nextIndex: 0,
      nextHandle: 1
    };
    Core.frozen = true;
  }
  fn();
  if (execute) {
    Core.frozen = false;
    processUpdates(Core.queues.current, true);
    Core.queues.current = prevQueue;
  }
};

var setContext = function(newContext, fn) {
  var context, ret;
  context = Core.context;
  Core.context = newContext;
  ret = fn();
  Core.context = context;
  return ret;
};

var root = function(fn) {
  var d;
  return setContext({
    disposables: d = []
  }, function() {
    return fn(function() {
      var disposable, i, len;
      for (i = 0, len = d.length; i < len; i++) {
        disposable = d[i];
        disposable();
      }
      d = [];
    });
  });
};

var onClean = function(fn) {
  if (!Core.context) {
    return;
  }
  return Core.context.disposables.push(fn);
};

var ignore = function(fn) {
  var disposables;
  ({disposables} = Core.context || {});
  return setContext({disposables}, fn);
};

var isObject = function(obj) {
  var ref;
  return obj !== null && ((ref = typeof obj) === 'object' || ref === 'function');
};

var isFunction = function(val) {
  return typeof val === 'function';
};

var diff = function(a, b, path = []) {
  var i, k, l, len, r, v;
  r = [];
  if (!isObject(a) || (b == null)) {
    if (a !== b) {
      r.push(path.concat([a]));
    }
  } else if (Array.isArray(a)) {
    for (k = i = 0, len = a.length; i < len; k = ++i) {
      v = a[k];
      if ((b != null ? b[k] : void 0) !== v) {
        comparer(v, k, b, true, path, r);
      }
    }
    if ((b != null ? b.length : void 0) > a.length) {
      l = a.length;
      while (l < b.length) {
        r.push(path.concat([l, void 0]));
        l++;
      }
    }
  } else {
    for (k in a) {
      v = a[k];
      if ((b != null ? b[k] : void 0) !== v) {
        comparer(v, k, b, false, path, r);
      }
    }
    for (k in b) {
      v = b[k];
      if (!(k in a)) {
        r.push(path.concat([k, void 0]));
      }
    }
  }
  return r;
};

var clone = function(v) {
  var k, obj;
  if (!isObject(v)) {
    return v;
  }
  if (Array.isArray(v)) {
    return v.slice(0);
  }
  obj = {};
  for (k in v) {
    obj[k] = v[k];
  }
  return obj;
};

var unwrap = function(item, deep) {
  var k, result, unwrapped, v;
  if (result = item != null ? item._target : void 0) {
    return result;
  }
  if (!(deep && isObject(item) && !isFunction(item) && !(item instanceof Element) && !(item instanceof DocumentFragment))) {
    return item;
  }
  for (k in item) {
    v = item[k];
    if ((unwrapped = unwrap(v, true)) !== v) {
      item[k] = unwrapped;
    }
  }
  return item;
};

var isEqual = function(a, b) {
  return ((a != null ? a._target : void 0) || a) === ((b != null ? b._target : void 0) || b);
};

var S, Selector, Signal, Stream, counter, fromObservable, fromPromise, resolveAsync, rxCompat, sequenceCounter;

counter = 0;

sequenceCounter = 0;

fromPromise = function(promise) {
  return {
    start: function(stream) {
      var complete;
      complete = false;
      return promise.then((value) => {
        if (complete) {
          return;
        }
        return stream._next(value);
      }).catch(err)(() => {
        return stream._error(err);
      });
    },
    stop: function() {
      var complete;
      return complete = true;
    }
  };
};

fromObservable = function(observable) {
  var disposable;
  disposable = null;
  return {
    start: function(stream) {
      return disposable = observable.subscribe((value) => {
        return stream._next(value);
      }, (error) => {
        return stream._error(error);
      }, () => {
        return stream._complete();
      });
    },
    stop: function() {
      disposable.unsubscribe();
      return disposable = null;
    }
  };
};

resolveAsync = function(value, seq, obsv, fn) {
  if (!isObject(value)) {
    return fn(value);
  }
  if ('subscribe' in value) {
    obsv.disposables.push(value.subscribe(fn));
    return;
  }
  if ('then' in value) {
    value.then(function(value) {
      if (seq < obsv.seq) {
        return;
      }
      return fn(value);
    });
    return;
  }
  return fn(value);
};

rxCompat = function(obsv, op) {
  var newObsv;
  newObsv = {
    subscribe: function(sub) {
      return op.call(sub, obsv);
    },
    lift: function(op) {
      return rxCompat(this, op);
    }
  };
  Object.defineProperty(newObsv, $$observable, {
    value: function() {
      return newObsv;
    }
  });
  return newObsv;
};

Signal = class Signal {
  constructor(value) {
    this.sid = `s_${counter++}`;
    this._subscriptions = new Set();
    this._value = value;
    Object.defineProperty(this, 'value', {
      get: function() {
        var ref;
        if ((ref = Core$1.context) != null ? ref.exec : void 0) {
          Core$1.context.disposables.push(this._subscribe(Core$1.context.exec).unsubscribe);
        }
        return this._value;
      },
      enumerable: true
    });
    Object.defineProperty(this, $$observable, {
      value: function() {
        return this;
      }
    });
  }

  peek() {
    return this._value;
  }

  next(value) {
    this._value = value;
    return this.notify('next', value);
  }

  error(value) {
    return this.notify('error', value);
  }

  complete() {
    return this.notify('complete');
  }

  subscribe(observer) {
    var d;
    d = this._subscribe.apply(this, arguments);
    if (this._value !== void 0) {
      (observer.next || observer).call(observer, this._value);
    }
    return d;
  }

  _subscribe(observer) {
    var onComplete;
    if (typeof observer !== 'object' || observer === null) {
      onComplete = arguments[2];
      observer = {
        next: observer,
        error: arguments[1],
        complete: () => {
          if (typeof onComplete === "function") {
            onComplete();
          }
          return this.unsubscribe(observer);
        }
      };
    }
    this._subscriptions.add(observer);
    return {
      unsubscribe: this.unsubscribe.bind(this, observer)
    };
  }

  unsubscribe(observer) {
    return this._subscriptions.delete(observer);
  }

  notify(type, value) {
    var handler, i, ref, size, sub;
    if (!(size = this._subscriptions.size)) {
      return;
    }
    i = 0;
    ref = this._subscriptions;
    for (sub of ref) {
      if (!(handler = sub[type])) {
        continue;
      }
      queueTask(handler.bind(sub), value);
      if (++i === size) {
        break;
      }
    }
  }

  map(fn) {
    return new Selector(() => {
      var v;
      if ((v = this.value) === void 0) {
        return;
      }
      return ignore(() => {
        return fn(v);
      });
    }, {
      notifyAlways: true
    });
  }

  memo(mapFn) {
    var disposables, length, list, mapped, signals, trackIndex;
    mapped = [];
    list = [];
    disposables = [];
    length = 0;
    if ((trackIndex = mapFn.length > 1)) {
      signals = [];
    }
    Core$1.context.disposables.push(function() {
      var d, k, len;
      for (k = 0, len = disposables.length; k < len; k++) {
        d = disposables[k];
        d();
      }
      return disposables = mapped = list = signals = null;
    });
    return this.map((newList) => {
      var d, end, i, indexedItems, item, itemIndex, j, k, l, len, len1, len2, m, mappedFn, newEnd, newLength, newListUnwrapped, newMapped, start, tempDisposables, tempSignals;
      // non-arrays
      newListUnwrapped = unwrap(newList, true);
      if (!Array.isArray(newListUnwrapped)) {
        if ((newListUnwrapped == null) || newListUnwrapped === false) {
          mapped = [];
          list = [];
          for (k = 0, len = disposables.length; k < len; k++) {
            d = disposables[k];
            d();
          }
          disposables = [];
          if (trackIndex) {
            signals = [];
          }
          return null;
        }
        if (list[0] === newListUnwrapped) {
          return mapped[0];
        }
        for (l = 0, len1 = disposables.length; l < len1; l++) {
          d = disposables[l];
          d();
        }
        disposables = [];
        list[0] = newListUnwrapped;
        return mapped[0] = root(function(dispose) {
          disposables[0] = dispose;
          return mapFn(newList);
        });
      }
      mappedFn = function(dispose) {
        disposables[j] = dispose;
        if (trackIndex) {
          return mapFn((typeof newList.peek === "function" ? newList.peek(j) : void 0) || newList[j], signals[j] = new Signal(j), newList);
        }
        return mapFn((typeof newList.peek === "function" ? newList.peek(j) : void 0) || newList[j]);
      };
      newLength = newListUnwrapped.length;
      if (newLength === 0) {
        if (length !== 0) {
          list = [];
          mapped = [];
          length = 0;
          for (m = 0, len2 = disposables.length; m < len2; m++) {
            d = disposables[m];
            d();
          }
          disposables = [];
          if (trackIndex) {
            signals = [];
          }
        }
      } else if (length === 0) {
        j = 0;
        while (j < newLength) {
          list[j] = newListUnwrapped[j];
          mapped[j] = root(mappedFn);
          j++;
        }
        length = newLength;
      } else {
        newMapped = new Array(newLength);
        tempDisposables = new Array(newLength);
        indexedItems = new Map();
        if (trackIndex) {
          tempSignals = new Array(newLength);
        }
        end = Math.min(length, newLength);
        // reduce from both ends
        start = 0;
        while (start < end && newListUnwrapped[start] === list[start]) {
          start++;
        }
        end = length - 1;
        newEnd = newLength - 1;
        while (end >= 0 && newEnd >= 0 && newListUnwrapped[newEnd] === list[end]) {
          newMapped[newEnd] = mapped[end];
          tempDisposables[newEnd] = disposables[end];
          if (trackIndex) {
            tempSignals[newEnd] = signals[end];
          }
          end--;
          newEnd--;
        }
        // create indices
        j = newEnd;
        while (j >= start) {
          item = newListUnwrapped[j];
          itemIndex = indexedItems.get(item);
          if (itemIndex != null) {
            itemIndex.push(j);
          } else {
            indexedItems.set(item, [j]);
          }
          j--;
        }
        // find old items
        i = start;
        while (i <= end) {
          item = list[i];
          itemIndex = indexedItems.get(item);
          if ((itemIndex != null) && itemIndex.length > 0) {
            j = itemIndex.pop();
            newMapped[j] = mapped[i];
            tempDisposables[j] = disposables[i];
            if (trackIndex) {
              tempSignals[j] = signals[i];
            }
          } else {
            disposables[i]();
          }
          i++;
        }
        // set all new values
        j = start;
        while (j < newLength) {
          if (newMapped.hasOwnProperty(j)) {
            mapped[j] = newMapped[j];
            disposables[j] = tempDisposables[j];
            if (trackIndex) {
              signals[j] = tempSignals[j];
              if (signals[j].peek() !== j) {
                signals[j].next(j);
              }
            }
          } else {
            mapped[j] = root(mappedFn);
          }
          j++;
        }
        // truncate extra length
        length = mapped.length = disposables.length = newLength;
        if (trackIndex) {
          signals.length = newLength;
        }
        // save list for next iteration
        list = newListUnwrapped.slice(0);
      }
      return mapped;
    });
  }

  // RxJS compatibility
  lift(op) {
    return rxCompat(this, op);
  }

  pipe() {
    return pipe.apply(this, arguments)(this);
  }

};

Stream = class Stream extends Signal {
  constructor(provider, {value}) {
    var ref;
    super(value);
    this.provider = provider;
    if ((ref = Core$1.context) != null) {
      ref.disposables.push(this.dispose.bind(this));
    }
  }

  _subscribe() {
    var disposable, oldSize;
    oldSize = this._subscriptions.size;
    disposable = super._subscribe.apply(this, arguments);
    if (!oldSize && this._subscriptions.size && !this.initialized) {
      this.provider.start(this);
      this.initialized = true;
    }
    return disposable;
  }

  unsubscribe(fn) {
    var delayed;
    super.unsubscribe(fn);
    delayed = () => {
      if (this.initialized && !this._subscriptions.size) {
        this.provider.stop(this);
        return delete this.initialized;
      }
    };
    delayed.defer = true;
    return queueTask(delayed);
  }

  next() {}

  error() {}

  complete() {}

  _next(value) {
    return super.next(value);
  }

  _error(value) {
    return super.error(value);
  }

  _complete() {
    return super.complete();
  }

  peek() {
    if (!this.initialized) {
      this.provider.start(this);
      this.initialized = true;
    }
    return super.peek();
  }

  dispose() {
    if (!this.initialized) {
      return;
    }
    this.provider.stop(this);
    return delete this.initialized;
  }

};

Selector = class Selector extends Signal {
  constructor(handler1, {defer, notifyAlways, value} = {}) {
    var ref;
    super(value);
    this.handler = handler1;
    this.notifyAlways = notifyAlways;
    this.exec = this.exec.bind(this);
    this.exec.defer = defer;
    this.disposables = [];
    this.pure = true;
    this.seq = 0;
    if ((ref = Core$1.context) != null) {
      ref.disposables.push(this.dispose.bind(this));
    }
  }

  exec() {
    var context, is_sleeping, seq, value;
    if (this.__disposed) {
      return;
    }
    this.clean();
    is_sleeping = !this._forceAwake && !this._subscriptions.size;
    context = is_sleeping ? null : this;
    value = setContext(context, () => {
      return this.handler(this._value);
    });
    seq = ++sequenceCounter;
    return resolveAsync(value, seq, this, (value) => {
      if (this.__disposed || value === void 0 || (!this.notifyAlways && value === this._value && !isObject(value))) {
        return;
      }
      this.seq = seq;
      return super.next(value);
    });
  }

  next() {}

  error() {}

  complete() {}

  peek() {
    if (!this.disposables.length) {
      this.exec();
    }
    return super.peek();
  }

  _subscribe() {
    if (!(this._subscriptions.size || this.disposables.length)) {
      this._forceAwake = true;
      this.exec();
      delete this._forceAwake;
    }
    return super._subscribe.apply(this, arguments);
  }

  unsubscribe(fn) {
    var delayed;
    super.unsubscribe(fn);
    delayed = () => {
      if (!this._subscriptions.size) {
        return this.clean();
      }
    };
    delayed.defer = true;
    return queueTask(delayed);
  }

  clean() {
    var disposable, k, len, ref;
    ref = this.disposables;
    for (k = 0, len = ref.length; k < len; k++) {
      disposable = ref[k];
      disposable();
    }
    return this.disposables = [];
  }

  dispose() {
    if (this.__disposed) {
      return;
    }
    this.notify('complete');
    this.clean();
    return this.__disposed = true;
  }

};

var Signal$1 = S = function(payload, options = {}) {
  if (!((payload != null) && isObject(payload))) {
    return new Signal(payload);
  }
  if ('sid' in payload) {
    return payload;
  }
  if (isFunction(payload)) {
    return new Selector(payload, options);
  }
  if ($$observable in payload) {
    return new Stream(fromObservable(payload[$$observable]()), options);
  }
  if ('then' in payload) {
    return new Stream(fromPromise(payload), options);
  }
  if ('start' in payload && 'stop' in payload) {
    return new Stream(payload, options);
  }
  return new Signal(payload);
};

var map = function(fn) {
  return function(obsv) {
    return S(obsv).map(fn);
  };
};

var memo = function(fn) {
  return function(obsv) {
    return S(obsv).memo(fn);
  };
};

var pipe = function(...fns) {
  if (!fns) {
    return i(() => {
      return i;
    });
  }
  if (fns.length === 1) {
    return fns[0];
  }
  return (obsv) => {
    obsv = S(obsv);
    return fns.reduce(((prev, fn) => {
      return fn(prev);
    }), obsv);
  };
};

var DEFINED, Handler;

DEFINED = {
  '_target': true,
  'on': true,
  'peek': true
};

var Handler$1 = Handler = class Handler {
  constructor(_target) {
    var i, len, method, ref;
    this._target = _target;
    this._child_subscriptions = {};
    ref = Object.keys(DEFINED).slice(2);
    for (i = 0, len = ref.length; i < len; i++) {
      method = ref[i];
      this[method] = this[method].bind(this);
    }
  }

  get(target, property) {
    var ref, value;
    if (DEFINED[property]) {
      return this[property];
    }
    if (!((ref = Core$1.context) != null ? ref.exec : void 0) || property === 'length' || typeof property === 'symbol') {
      return target[property];
    }
    if ((value = target[property]) != null) {
      if (isFunction(value)) {
        return value;
      }
      if (isObject(value) && Object.isFrozen(value)) {
        value = clone(value);
        target[property] = value;
      }
      value = Handler.wrap(value);
    }
    Core$1.context.disposables.push(this.on(property, Core$1.context.exec).unsubscribe);
    return value;
  }

  set(target, property, value) {
    return true;
  }

  deleteProperty(target, property) {
    return true;
  }

  subscribe(fn) {
    this.__subscriptions || (this.__subscriptions = new Set());
    this.__subscriptions.add(fn);
    return {
      unsubscribe: () => {
        return this.__subscriptions.delete(fn);
      }
    };
  }

  notify(value) {
    var ref, ref1, sub;
    if (!((ref = this.__subscriptions) != null ? ref.size : void 0)) {
      return;
    }
    ref1 = this.__subscriptions;
    for (sub of ref1) {
      queueTask(sub, value);
    }
  }

};

Handler.wrappers = new WeakMap();

Handler.wrap = function(value) {
  var handler, proxy, wrapper;
  if (!(isObject(value) && !(value instanceof Element))) {
    return value;
  }
  if (!(wrapper = Handler.wrappers.get(value))) {
    proxy = new Proxy(value, handler = new Handler(value));
    Handler.wrappers.set(value, {proxy, handler});
  } else {
    ({proxy} = wrapper);
  }
  return proxy;
};

Handler.handler = function(value) {
  var handler, proxy, wrapper;
  if (!(isObject(value) && !(value instanceof Element))) {
    return value;
  }
  if (!(wrapper = Handler.wrappers.get(value))) {
    proxy = new Proxy(value, handler = new Handler(value));
    Handler.wrappers.set(value, {proxy, handler});
  } else {
    ({handler} = wrapper);
  }
  return handler;
};

var methods;

methods = {
  trigger: function(property, value, notify) {
    var sub, subs;
    subs = this._child_subscriptions[property];
    if (subs != null ? subs.size : void 0) {
      for (sub of subs) {
        queueTask(sub, value);
      }
    }
    if (notify) {
      return this.notify(this._target);
    }
  },
  peek: function(property) {
    var value;
    value = this._target[property];
    if (isObject(value) && Object.isFrozen(value)) {
      value = clone(value);
      this._target[property] = value;
    }
    return Handler$1.wrap(value);
  },
  on: function(property, fn) {
    var base, disposable, handler, value;
    value = this._target[property];
    disposable = null;
    if (isObject(value) && !(isFunction(value) || value instanceof Element)) {
      handler = Handler$1.handler(value);
      disposable = handler.subscribe(fn);
    }
    (base = this._child_subscriptions)[property] || (base[property] = new Set());
    this._child_subscriptions[property].add(fn);
    return {
      unsubscribe: () => {
        if (disposable) {
          disposable.unsubscribe();
        }
        return this._child_subscriptions[property].delete(fn);
      }
    };
  }
};

Object.assign(Handler$1.prototype, methods);

var methods$1 = methods;

var State, setNested;

setNested = function(item, changes) {
  var handler, isArray, notify, property, value;
  handler = Handler$1.handler(item);
  isArray = Array.isArray(item);
  if (arguments.length === 3) {
    notify = isArray || !(arguments[1] in item);
    if (arguments[2] === void 0) {
      delete item[arguments[1]];
      if (isArray) {
        item.length--;
      }
    } else {
      item[arguments[1]] = arguments[2];
    }
    handler.trigger(arguments[1], arguments[2], notify);
    return;
  }
  for (property in changes) {
    value = changes[property];
    notify = isArray || !(property in item);
    if (value === void 0) {
      delete item[property];
    } else {
      item[property] = value;
    }
    handler.trigger(property, value, notify);
  }
};

var index = State = class State {
  constructor(state) {
    var k, ref;
    Object.defineProperties(this, {
      _target: {
        value: state,
        writable: true
      },
      _disposables: {
        value: [],
        writable: true
      },
      _child_subscriptions: {
        value: {},
        writable: true
      }
    });
    for (k in this._target) {
      this._defineProperty(k);
    }
    if ((ref = Core$1.context) != null) {
      ref.disposables.push(this.dispose.bind(this));
    }
  }

  set() {
    var args, ref;
    if ((ref = Core$1.context) != null ? ref.pure : void 0) {
      return console.log('Cannot update in a Selector');
    }
    args = arguments;
    run(() => {
      var change, changes, current, i, j, len, property, ref1, ref2, temp, value;
      if (args.length === 1) {
        if (Array.isArray(args[0])) {
          ref1 = args[0];
          for (j = 0, len = ref1.length; j < len; j++) {
            change = ref1[j];
            this.set.apply(this, change);
          }
        } else {
          ref2 = args[0];
          for (property in ref2) {
            value = ref2[property];
            this._setProperty(property, value);
          }
        }
        return;
      }
      current = this._target;
      changes = args[args.length - 1];
      i = 0;
      while (i < args.length - 1 && ((temp = current[args[i]]) != null)) {
        current = temp;
        i++;
      }
      return setNested(current, changes);
    });
  }

  replace() {
    var changes, current, i, property, temp, value;
    if (arguments.length === 1) {
      if (!(arguments[0] instanceof Object)) {
        return console.log('replace must be provided a replacement state');
      }
      changes = arguments[0];
      run(() => {
        var change, j, len;
        if (!Array.isArray(changes)) {
          changes = diff(changes, this._target);
        }
        for (j = 0, len = changes.length; j < len; j++) {
          change = changes[j];
          this.replace.apply(this, change);
        }
      });
    }
    if (arguments.length === 2) {
      this._setProperty(arguments[0], arguments[1]);
      return;
    }
    current = this._target;
    value = arguments[arguments.length - 1];
    property = arguments[arguments.length - 2];
    i = 0;
    while (i < arguments.length - 2 && ((temp = current[arguments[i]]) != null)) {
      current = temp;
      i++;
    }
    setNested(current, property, value);
  }

  select() {
    var j, key, len, selection, selector;
    for (j = 0, len = arguments.length; j < len; j++) {
      selection = arguments[j];
      if (isFunction(selection) || 'then' in selection || 'subscribe' in selection) {
        if (!('sid' in selection)) {
          selection = Signal$1(selection);
        }
        this._disposables.push(selection.subscribe((value) => {
          var key, val;
          value = unwrap(value, true);
          this.replace([].concat(...((function() {
            var ref, results;
            ref = value || {};
            results = [];
            for (key in ref) {
              val = ref[key];
              results.push(diff(val, this._target[key], [key]));
            }
            return results;
          }).call(this))));
        }));
        return;
      }
      for (key in selection) {
        selector = selection[key];
        ((key, selector) => {
          if (!(key in this)) {
            this._defineProperty(key);
          }
          if (!('sid' in selector)) {
            selector = Signal$1(selector);
          }
          this._disposables.push(selector.subscribe((value) => {
            this.replace(diff(unwrap(value, true), this._target[key], [key]));
          }));
        })(key, selector);
      }
    }
  }

  dispose() {
    var disposable, j, len, ref;
    if (!this._disposables) {
      return;
    }
    ref = this._disposables;
    for (j = 0, len = ref.length; j < len; j++) {
      disposable = ref[j];
      disposable.unsubscribe();
    }
    this._disposables = null;
  }

  _setProperty(property, value) {
    if (!(property in this)) {
      this._defineProperty(property);
    }
    if (value === void 0) {
      delete this._target[property];
    } else {
      this._target[property] = value;
    }
    return this.trigger(property, value);
  }

  _defineProperty(property) {
    return Object.defineProperty(this, property, {
      get: () => {
        var ref, value;
        if (!((ref = Core$1.context) != null ? ref.exec : void 0)) {
          return this._target[property];
        }
        if ((value = this._target[property]) != null) {
          if (isFunction(value)) {
            return value;
          }
          if (isObject(value) && Object.isFrozen(value)) {
            value = clone(value);
            this._target[property] = value;
          }
          value = Handler$1.wrap(value);
        }
        Core$1.context.disposables.push(this.on(property, Core$1.context.exec).unsubscribe);
        return value;
      },
      enumerable: true
    });
  }

};

Object.assign(State.prototype, methods$1);

var DEFINED$1, ImmutableHandler;

DEFINED$1 = {
  '_target': true,
  'on': true,
  'peek': true,
  'map': true
};

var ImmutableHandler$1 = ImmutableHandler = class ImmutableHandler {
  constructor(target, _target, root$$1) {
    this._target = _target;
    this.root = root$$1;
    this.peek = this.peek.bind(this, target);
    this.on = this.on.bind(this, target);
    if (Array.isArray(this._target)) {
      this.map = this.map.bind(this, target);
    }
    this.stateClock = target.clock || Core$1.clock;
  }

  get(target, property) {
    var base, base1, ref, value;
    if (this.stateClock < target.clock) {
      this.resolveState(target);
    }
    if (property === 'length' || typeof property === 'symbol') {
      return this._target[property];
    }
    if (property.endsWith('$')) {
      property = property.slice(0, -1);
      if (Core$1.context.exec) {
        target[property] || (target[property] = {
          _subs: new Set(),
          clock: Core$1.clock
        });
        (base = target[property])._subPath || (base._subPath = new Set());
        target[property]._subPath.add(Core$1.context.exec);
        Core$1.context.disposables.push(target[property]._subPath.delete(Core$1.context.exec));
      }
      return this._target[property];
    }
    if (DEFINED$1[property]) {
      return this[property];
    }
    if (!((ref = Core$1.context) != null ? ref.exec : void 0)) {
      return this._target[property];
    }
    if (((value = this._target[property]) != null) && isObject(value) && !(value instanceof Element)) {
      if (isFunction(value)) {
        return value;
      }
      target[property] || (target[property] = {
        clock: Core$1.clock
      });
      (base1 = target[property]).path || (base1.path = target.path.concat([property]));
      value = new Proxy(target[property], new ImmutableHandler(target[property], value, this.root));
    }
    Core$1.context.disposables.push(this.on(property, Core$1.context.exec).unsubscribe);
    return value;
  }

  set(target, property, value) {
    return true;
  }

  deleteProperty(target, property) {
    return true;
  }

  on(target, property, fn) {
    var base;
    target[property] || (target[property] = {
      clock: Core$1.clock
    });
    (base = target[property])._subs || (base._subs = new Set());
    target[property]._subs.add(fn);
    return {
      unsubscribe: function() {
        return target[property]._subs.delete(fn);
      }
    };
  }

  peek(target, property) {
    var base, value;
    if (this.stateClock < target.clock) {
      this.resolveState(target);
    }
    value = this._target[property];
    if ((value == null) || isFunction(value) || !isObject(value) || (value instanceof Element)) {
      return value;
    }
    target[property] || (target[property] = {});
    (base = target[property]).path || (base.path = target.path.concat([property]));
    return new Proxy(target[property], new ImmutableHandler(target[property], value, this.root));
  }

  map(target, fn) {
    if (this.stateClock < target.clock) {
      this.resolveState(target);
    }
    return this._target.map((value, i, a) => {
      var base;
      if (!((value == null) || isFunction(value) || !isObject(value) || (value instanceof Element))) {
        target[i] || (target[i] = {});
        (base = target[i]).path || (base.path = target.path.concat([i]));
        value = new Proxy(target[i], new ImmutableHandler(target[i], value, this.root));
      }
      return fn(value, i, a);
    });
  }

  resolveState(target) {
    var current, i, l;
    current = this.root._target;
    i = 0;
    l = target.path.length;
    while (i < l) {
      if (current) {
        current = current[target.path[i]];
      }
      i++;
    }
    this._target = current;
    return this.stateClock = Core$1.clock;
  }

};

var ImmutableState,
  indexOf = [].indexOf;

var index$1 = ImmutableState = class ImmutableState {
  constructor(state) {
    var k, ref;
    Object.defineProperties(this, {
      _target: {
        value: state,
        writable: true
      },
      _disposables: {
        value: [],
        writable: true
      },
      _subscriptions: {
        value: {},
        writable: true
      }
    });
    for (k in this._target) {
      this._defineProperty(k);
    }
    if ((ref = Core$1.context) != null) {
      ref.disposables.push(this.dispose.bind(this));
    }
  }

  set() {
    var args, ref;
    if ((ref = Core$1.context) != null ? ref.pure : void 0) {
      return console.log('Cannot update in a Selector');
    }
    args = arguments;
    return run(() => {
      var change, changes, j, k, l, len, len1, notify, path, property, ref1, ref2, ref3, state, subKeys, subPaths, subs, value;
      if (args.length === 1) {
        if (Array.isArray(args[0])) {
          ref1 = args[0];
          for (j = 0, len = ref1.length; j < len; j++) {
            change = ref1[j];
            this.set.apply(this, change);
          }
        } else {
          ref2 = args[0];
          for (property in ref2) {
            value = ref2[property];
            this._setProperty(property, value);
          }
        }
        return;
      }
      changes = args[args.length - 1];
      ({state, subs, subPaths} = this._resolvePath(args, args.length - 1));
      if (!state) {
        return;
      }
      notify = Array.isArray(state);
      for (property in changes) {
        value = changes[property];
        notify = notify || !(indexOf.call(state, property) >= 0);
        if (value === void 0) {
          delete state[property];
        } else {
          state[property] = value;
        }
        if (subs) {
          this.trigger(subs != null ? (ref3 = subs[property]) != null ? ref3._subs : void 0 : void 0, value);
        }
      }
      if (notify) {
        this.trigger(subs != null ? subs._subs : void 0, state);
      }
      for (l = 0, len1 = subPaths.length; l < len1; l++) {
        path = subPaths[l];
        this.trigger(path.subs, path.state);
      }
      // prune subs
      if (value === void 0) {
        delete subs[property];
      } else if (Array.isArray(value) && !value.length) {
        subKeys = ['_subs', 'clock', '_subPath'];
        for (k in subs[property]) {
          if (!(indexOf.call(subKeys, k) >= 0)) {
            delete subs[property][k];
          }
        }
      }
    });
  }

  replace() {
    var changes, j, len, notify, path, property, ref, state, subPaths, subs, value;
    if (arguments.length === 1) {
      if (!(arguments[0] instanceof Object)) {
        return console.log('replace must be provided a replacement state');
      }
      changes = arguments[0];
      run(() => {
        var change, j, len;
        if (!Array.isArray(changes)) {
          changes = diff(changes, this._target);
        }
        for (j = 0, len = changes.length; j < len; j++) {
          change = changes[j];
          this.replace.apply(this, change);
        }
      });
      return;
    }
    if (arguments.length === 2) {
      this._setProperty(arguments[0], arguments[1]);
      return;
    }
    value = arguments[arguments.length - 1];
    property = arguments[arguments.length - 2];
    ({state, subs, subPaths} = this._resolvePath(arguments, arguments.length - 2));
    if (!state) {
      return;
    }
    if (value === void 0 && Array.isArray(state)) {
      state.length -= 1;
      notify = true;
    } else {
      notify = Array.isArray(state) || !(indexOf.call(state, property) >= 0);
      if (value === void 0) {
        delete state[property];
      } else {
        state[property] = value;
      }
    }
    if (subs) {
      this.trigger((ref = subs[property]) != null ? ref._subs : void 0, value);
    }
    if (notify) {
      this.trigger(subs._subs, state);
    }
    for (j = 0, len = subPaths.length; j < len; j++) {
      path = subPaths[j];
      this.trigger(path.subs, path.state);
    }
  }

  select() {
    var j, key, len, selection, selector;
    for (j = 0, len = arguments.length; j < len; j++) {
      selection = arguments[j];
      if (isFunction(selection) || 'then' in selection || 'subscribe' in selection) {
        if (!('sid' in selection)) {
          selection = Signal$1(selection);
        }
        this._disposables.push(selection.subscribe((value) => {
          var key, val;
          value = unwrap(value, true);
          this.replace([].concat(...((function() {
            var ref, results;
            ref = value || {};
            results = [];
            for (key in ref) {
              val = ref[key];
              results.push(diff(val, this._target[key], [key]));
            }
            return results;
          }).call(this))));
        }));
        return;
      }
      for (key in selection) {
        selector = selection[key];
        ((key, selector) => {
          if (!(key in this)) {
            this._defineProperty(key);
          }
          if (!('sid' in selector)) {
            selector = Signal$1(selector);
          }
          this._disposables.push(selector.subscribe((value) => {
            this.replace(diff(unwrap(value, true), this._target[key], [key]));
          }));
        })(key, selector);
      }
    }
  }

  on(property, fn) {
    var base, base1;
    (base = this._subscriptions)[property] || (base[property] = {
      clock: Core$1.clock
    });
    (base1 = this._subscriptions[property])._subs || (base1._subs = new Set());
    this._subscriptions[property]._subs.add(fn);
    return {
      unsubscribe: () => {
        return this._subscriptions[property]._subs.delete(fn);
      }
    };
  }

  trigger(subs, value) {
    var sub;
    if (!(subs != null ? subs.size : void 0)) {
      return;
    }
    for (sub of subs) {
      queueTask(sub, value);
    }
  }

  dispose() {
    var disposable, j, len, ref;
    if (!this._disposables) {
      return;
    }
    ref = this._disposables;
    for (j = 0, len = ref.length; j < len; j++) {
      disposable = ref[j];
      disposable.unsubscribe();
    }
    this._disposables = null;
  }

  _resolvePath(path, length) {
    var currentState, currentSubs, i, ref, subPaths;
    currentState = this._target;
    currentSubs = this._subscriptions;
    subPaths = [];
    i = 0;
    while (i < length) {
      if (!currentSubs[path[i]]) {
        currentSubs[path[i]] = {};
      }
      currentSubs = currentSubs[path[i]];
      if (currentState != null) {
        if (currentSubs.clock !== Core$1.clock) {
          currentState[path[i]] = clone(currentState[path[i]]);
          currentSubs.clock = Core$1.clock;
        }
        currentState = currentState[path[i]];
      }
      if (currentSubs != null ? (ref = currentSubs._subPath) != null ? ref.size : void 0 : void 0) {
        subPaths.push({
          state: currentState,
          subs: currentSubs._subPath
        });
      }
      i++;
    }
    return {
      state: currentState,
      subs: currentSubs,
      subPaths
    };
  }

  _setProperty(property, value) {
    var k, ref, ref1, subKeys;
    if (!(property in this)) {
      this._defineProperty(property);
    }
    if (value === void 0) {
      delete this._target[property];
    } else {
      this._target[property] = value;
    }
    this.trigger((ref = this._subscriptions[property]) != null ? ref._subs : void 0, value);
    this.trigger((ref1 = this._subscriptions[property]) != null ? ref1._subPath : void 0, value);
    // prune subs
    if (value === void 0) {
      delete this._subscriptions[property];
    } else if (Array.isArray(value) && !value.length) {
      subKeys = ['_subs', 'clock', '_subPath'];
      for (k in this._subscriptions[property]) {
        if (!(indexOf.call(subKeys, k) >= 0)) {
          delete this._subscriptions[property][k];
        }
      }
    }
  }

  _defineProperty(property) {
    Object.defineProperty(this, property, {
      get: () => {
        var base, base1, ref, value;
        if (!((ref = Core$1.context) != null ? ref.exec : void 0)) {
          return this._target[property];
        }
        if (((value = this._target[property]) != null) && isObject(value) && !(value instanceof Element)) {
          if (isFunction(value)) {
            return value;
          }
          (base = this._subscriptions)[property] || (base[property] = {
            clock: Core$1.clock
          });
          (base1 = this._subscriptions[property]).path || (base1.path = [property]);
          value = new Proxy(this._subscriptions[property], new ImmutableHandler$1(this._subscriptions[property], value, this));
        }
        Core$1.context.disposables.push(this.on(property, Core$1.context.exec).unsubscribe);
        return value;
      },
      enumerable: true
    });
    return Object.defineProperty(this, property + 'Signal', {
      get: () => {
        var base, base1, fn;
        if (fn = Core$1.context.exec) {
          (base = this._subscriptions)[property] || (base[property] = {
            _subs: new Set(),
            clock: Core$1.clock
          });
          (base1 = this._subscriptions[property])._subPath || (base1._subPath = new Set());
          this._subscriptions[property]._subPath.add(Core$1.context.exec);
          Core$1.context.disposables.push(() => {
            return this._subscriptions[property]._subPath.delete(fn);
          });
        }
        return this._target[property];
      }
    });
  }

};

var Sync;

var Sync$1 = Sync = class Sync {
  constructor(handler, {defer = true} = {}) {
    var ref;
    this.handler = handler;
    this.exec = this.exec.bind(this);
    this.exec.defer = defer;
    this.disposables = [];
    this.exec();
    if ((ref = Core$1.context) != null) {
      ref.disposables.push(this.dispose.bind(this));
    }
  }

  exec() {
    if (this.__disposed) {
      return;
    }
    this.clean();
    return setContext(this, this.handler);
  }

  clean() {
    var disposable, i, len, ref;
    ref = this.disposables;
    for (i = 0, len = ref.length; i < len; i++) {
      disposable = ref[i];
      disposable();
    }
    return this.disposables = [];
  }

  dispose() {
    if (this.__disposed) {
      return;
    }
    this.clean();
    return this.__disposed = true;
  }

};

export { index as State, index$1 as ImmutableState, Sync$1 as Sync, ignore, root, unwrap, isEqual, onClean, Signal$1 as S, map, memo, pipe };
